#include <conio.h>
#include <ctype.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>

#include "emu.h"

/* cycle_count_delta is a (usually negative) number telling what the time is relative
 * to the next timer interrupt that will be executed.
 * (cycle_count + cycle_count_delta) is the total number of cycles executed so far */
u64 cycle_count = 0;
int cycle_count_delta = 0;

int throttle_delay = 10; /* in milliseconds */

u32 cpu_events;

bool exiting;
bool do_translate = true;
bool emulate_cas;

bool turbo_mode;
bool is_halting;
bool show_speed;

jmp_buf restart_after_exception;

const char log_type_tbl[] = LOG_TYPE_TBL;
int log_enabled[MAX_LOG];
FILE *log_file[MAX_LOG];
void logprintf(int type, char *str, ...) {
	if (log_enabled[type]) {
		va_list va;
		va_start(va, str);
		vfprintf(log_file[type], str, va);
		va_end(va);
	}
}

void warn(char *fmt, ...) {
	va_list va;
	fprintf(stderr, "Warning at PC=%08X: ", arm.reg[15]);
	va_start(va, fmt);
	vfprintf(stderr, fmt, va);
	va_end(va);
	fprintf(stderr, "\n");
}

__attribute__((noreturn))
void error(char *fmt, ...) {
	va_list va;
	fprintf(stderr, "Error at PC=%08X: ", arm.reg[15]);
	va_start(va, fmt);
	vfprintf(stderr, fmt, va);
	va_end(va);
	fprintf(stderr, "\n\tBacktrace:\n");
	backtrace(arm.reg[11]);
	debugger();
	cpu_events |= EVENT_RESET;
	longjmp(restart_after_exception, 0);
}

HANDLE hTimerEvent;
UINT uTimerID;
void throttle_timer_on() {
	uTimerID = timeSetEvent(throttle_delay, throttle_delay, (LPTIMECALLBACK)hTimerEvent, 0,
	                        TIME_PERIODIC | TIME_CALLBACK_EVENT_SET);
	if (uTimerID == 0) {
		printf("timeSetEvent failed\n");
		exit(1);
	}
}
void throttle_timer_off() {
	if (uTimerID != 0)
		timeKillEvent(uTimerID);
	uTimerID = 0;
}

int exec_hack() {
	return 0;
}

void prefetch_abort(u32 mva, u8 status) {
	logprintf(LOG_CPU, "Prefetch abort: address=%08x status=%02x\n", mva, status);
	arm.reg[15] += 4;
	// Fault address register not changed
	arm.instruction_fault_status = status;
	cpu_exception(EX_PREFETCH_ABORT);
	if (mva == arm.reg[15])
		error("Abort occurred with exception vectors unmapped");
	longjmp(restart_after_exception, 0);
}

void data_abort(u32 mva, u8 status) {
	logprintf(LOG_CPU, "Data abort: address=%08x status=%02x\n", mva, status);
	fix_pc_for_fault();
	arm.reg[15] += 8;
	arm.fault_address = mva;
	arm.data_fault_status = status;
	cpu_exception(EX_DATA_ABORT);
	longjmp(restart_after_exception, 0);
}

int main(int argc, char **argv) {
	int i;
	FILE *f;
	static FILE *boot2_file = NULL;
	static char *boot1_filename = NULL, *boot2_filename = NULL, *flash_filename = NULL;
	bool new_flash_image = false;
	char *preload_boot2 = NULL, *preload_diags = NULL, *preload_os = NULL;

	for (i = 1; i < argc; i++) {
		char *arg = argv[i];
		if (*arg == '/' || *arg == '-') {
			arg++;
			switch (toupper(*arg++)) {
				case '1':
					if (*arg == '=') arg++;
					boot1_filename = arg;
					break;
				case 'B':
					if (*arg == '=') arg++;
					boot2_filename = arg;
					boot2_file = fopen(boot2_filename, "rb");
					if (!boot2_file) {
						perror(boot2_filename);
						return 1;
					}
					break;
				case 'C':
					if (*arg) goto usage;
					emulate_cas = true;
					break;
				case 'D':
					if (*arg) goto usage;
					cpu_events |= EVENT_DEBUG_STEP;
					break;
				case 'F':
					if (*arg == '=') arg++;
					flash_filename = arg;
					break;
				case 'K':
					keypad_type = 1;
					if (*arg) {
						keypad_type = atoi(arg) - 1;
						if (keypad_type < 0 || keypad_type >= NUM_KEYPAD_TYPES)
							goto usage;
					}
					break;
				case 'L': {
					char *p = strchr(log_type_tbl, toupper(*arg++));
					int type = p - log_type_tbl;
					if (!p) goto usage;
					log_enabled[type] = 1;
					if (arg[0] == '=') {
						char *p2 = strchr(log_type_tbl, toupper(arg[1]));
						int type2 = p2 - log_type_tbl;
						if (!p2) goto usage;
						log_file[type] = log_file[type2];
					} else if (strcmp(arg, "-") == 0) {
						log_file[type] = stdout;
					} else {
						log_file[type] = fopen(arg, "wb");
					}
					if (!log_file[type]) {
						perror(arg);
						exit(1);
					}
					break;
				}
				case 'N':
					if (*arg) goto usage;
					new_flash_image = true;
					break;
				case 'P': {
					char **pp;
					switch (toupper(*arg++)) {
						case 'B': pp = &preload_boot2; break;
						case 'D': pp = &preload_diags; break;
						case 'O': pp = &preload_os; break;
						default: goto usage;
					}
					if (*arg == '=') arg++;
					*pp = arg;
					break;
				}
				default:
usage:
					printf(
						"nspire emulator v0.50\n"
						"  /1=boot1	- location of BOOT1 image\n"
						"  /B=boot2	- location of decompressed BOOT2 image\n"
						"  /C		- emulate CAS hardware version\n"
						"  /D		- enter debugger at start\n"
						"  /F=file	- flash image filename\n"
						"  /Kn		- set keypad type (2 = TI-84 Plus, 4 = Touchpad)\n"
						"  /N		- create new flash image\n"
						"  /PB=boot2.img	- preload flash with BOOT2 (.img file)\n"
						"  /PD=diags.img	- preload flash with DIAGS image\n"
						"  /PO=osfile	- preload flash with OS (.tnc/.tno file)\n");
					return 1;
			}
		} else {
			goto usage;
		}
	}

	switch ((boot1_filename != NULL) + (boot2_filename != NULL) + new_flash_image) {
		case 0: goto usage;
		case 1: break;
		default: printf("Must use exactly one of /1, /B, or /N.\n"); return 0;
	}

	if (new_flash_image) {
		flash_initialize(preload_boot2, preload_diags, preload_os);
		return flash_save_as(flash_filename);
	}

	if (flash_filename) {
		if (preload_boot2 || preload_diags || preload_os) {
			printf("Can't preload to an existing flash image\n");
			return 1;
		}
		flash_file = fopen(flash_filename, "r+b");
		if (!flash_file) {
			perror(flash_filename);
			return 1;
		}
		if (!fread(flash_data, sizeof flash_data, 1, flash_file)) {
			printf("Could not read flash image from %s\n", flash_filename);
			return 1;
		}
	} else {
		flash_initialize(preload_boot2, preload_diags, preload_os);
	}

	memset(MEM_PTR(0x00000000), -1, 0x80000);
	for (i = 0x00000; i < 0x80000; i += 4) {
		RAM_FLAGS(&rom_00[i]) = RF_READ_ONLY;
	}
	if (boot1_filename) {
		/* Load the ROM */
		f = fopen(boot1_filename, "rb");
		if (!f) {
			perror(boot1_filename);
			return 1;
		}
		fread(MEM_PTR(0x00000000), 1, 0x80000, f);
		fclose(f);
	} else {
		/* Little hack to get OS 1.1 to work */
		*(u32 *)MEM_PTR(0x0000001C) = 0x1234;
		*(u32 *)MEM_PTR(0x00001234) = 0xE12FFF1E; /* bx lr */
	}

	insn_buffer = VirtualAlloc(NULL, INSN_BUFFER_SIZE, MEM_RESERVE | MEM_COMMIT, PAGE_EXECUTE_READWRITE);
	insn_bufptr = insn_buffer;

	memory_initialize();
	addr_cache_init();

	LARGE_INTEGER perffreq;
	QueryPerformanceFrequency(&perffreq);

	hTimerEvent = CreateEvent(NULL, FALSE, FALSE, NULL);

	gui_initialize();

	throttle_timer_on();
	atexit(throttle_timer_off);

reset:
	memset(&arm, 0, sizeof arm);
	arm.cpsr_low28 = MODE_SVC | 0xC0;
	cpu_events &= EVENT_DEBUG_STEP;
	if (boot2_file) {
		/* Start from BOOT2. (needs to be re-loaded on each reset since
		 * it can get overwritten in memory) */
		fseek(boot2_file, 0, SEEK_SET);
		fread(MEM_PTR(0x11800000), 1, 0x800000, boot2_file);
		arm.reg[15] = 0x11800000;
		if (*(u8 *)MEM_PTR(0x11800003) < 0xE0) {
			printf("%s does not appear to be an uncompressed BOOT2 image.\n", boot2_filename);
			return 1;
		}

		/* To enter maintenance mode (home+enter+P), address A4012ECC
		 * must contain an array indicating those keys before BOOT2 starts */
		memcpy(MEM_PTR(0xA4012ECC), (void *)key_map, 0x12);

		/* Disable all FIQs (since BOOT2 neglects to do this) */
		intr.mask[1] = 0;
	} else {
		/* Start from BOOT1. */
		arm.reg[15] = 0;
	}
	arm.control = 0x00050078;
	addr_cache_flush();
	flush_translations();

	intr.noninverted = -1;
	intr.priority_limit[0] = 8;
	intr.priority_limit[1] = 8;

	gpio_reset();
	keypad_reset();

	/* Halt the hardware timers */
	for (i = 0; i < 3; i++) {
		timerpairs[i].timers[0].control = 0x10;
		timerpairs[i].timers[1].control = 0x10;
	}

	usb_reset();
	usblink_reset();

	setjmp(restart_after_exception);

	while (!exiting) {
		while (cycle_count_delta < 0) {
			if (cpu_events & EVENT_RESET) {
				printf("CPU reset\n");
				goto reset;
			}

			if (cpu_events & (EVENT_FIQ | EVENT_IRQ)) {
				if (cpu_events & EVENT_WAITING)
					arm.reg[15] += 4; // Skip over wait instruction
				logprintf(LOG_INTS, "Dispatching an interrupt\n");
				arm.reg[15] += 4;
				cpu_exception((cpu_events & EVENT_FIQ) ? EX_FIQ : EX_IRQ);
			}
			cpu_events &= ~EVENT_WAITING;

			if (arm.cpsr_low28 & 0x20)
				cpu_thumb_loop();
			else
				cpu_arm_loop();
		}

		// Timer ticks are approx. 32 kHz
		cycle_count       += 2812;
		cycle_count_delta -= 2812;

		timer_advance(&timerpairs[0], 703);
		timer_advance(&timerpairs[1], 1);
		timer_advance(&timerpairs[2], 1);

		keypad_scan_next();

		static int interval_ticks;
		if (--interval_ticks < 0) {
			/* Throttle interval (defined arbitrarily as 100Hz) - used for
			 * keeping the emulator speed down, and other miscellaneous stuff
			 * that needs to be done periodically */
			static int intervals;
			intervals++;
			interval_ticks += 320;

			extern void usblink_timer();
			usblink_timer();

			if (_kbhit()) {
				char c = _getch();
				if (c == 4)
					debugger();
				else
					serial_byte_in(c);
			}

			get_messages();

			LARGE_INTEGER interval_end;
			QueryPerformanceCounter(&interval_end);

			{	// Update graphics (frame rate is arbitrary)
				static LARGE_INTEGER prev;
				LONGLONG time = interval_end.QuadPart - prev.QuadPart;
				if (time >= (perffreq.QuadPart >> 5)) {
					InvalidateRect(hwndGfx, NULL, FALSE);
					prev = interval_end;
				}
			}

			if (show_speed) {
				// Show speed
				static int prev_intervals;
				static LARGE_INTEGER prev;
				LONGLONG time = interval_end.QuadPart - prev.QuadPart;
				if (time >= (perffreq.QuadPart >> 1)) {
					double speed = (double)perffreq.QuadPart * (intervals - prev_intervals) / time;
					char buf[40];
					sprintf(buf, "nspire_emu - %.1f%%", speed);
					SetWindowText(hwndMain, buf);
					prev_intervals = intervals;
					prev = interval_end;
				}
			}

			if (!turbo_mode || is_halting)
				WaitForSingleObject(hTimerEvent, INFINITE);
			if (is_halting)
				is_halting--;

#if 0
			if (log_enabled[LOG_ICOUNT]) {
				static LARGE_INTEGER interval_start;
				LARGE_INTEGER next_start;
				QueryPerformanceCounter(&next_start);

				static u64 prev_cycles;
				u64 cycles = cycle_count + cycle_count_delta;
				logprintf(LOG_ICOUNT, "Time=%7d (CPU=%7d Idle=%7d) Insns=%8d Rate=%9I64d\n",
					next_start.LowPart - interval_start.LowPart,
					interval_end.LowPart - interval_start.LowPart,
					next_start.LowPart - interval_end.LowPart,
					(u32)(cycles - prev_cycles),
					(cycles - prev_cycles) * perffreq.QuadPart / (interval_end.LowPart - interval_start.LowPart));
				prev_cycles = cycles;
				interval_start = next_start;
			}
#endif
		}
	}
	return 0;
}
